
#!/usr/bin/perl -Tw
# -*- perl -*-

use strict;
use CGI qw(:standard *table *Tr *dl);
use DBI;

use constant REGISTRATION_CODE_LENGTH => 100;
use vars qw(@CANDIDATES @PARTIES @OFFICES @CANDIDATE_NAME);
$ENV{PATH} = '/bin';

# connect to database
my $DB = DBI->connect('dbi:mysql:CEA') or die "Can't connect: $DBI::errstr";

########################
# create the page
########################
print header,
  start_html(-title   => 'Indecision 2000',
             -bgcolor => 'white'
            ),
  h1({-align=>'CENTER'},
     img({-src=>'/icons/star.gif'}) x 3,
     'Indecision 2000',
     img({-src=>'/icons/star.gif'}) x 3);

# load global variables
get_globals();

# If the VOTE button is pressed, validate and enter the ballot
if (param('VOTE') && (my $registration = validate())) {
    enter_ballot($registration);
}

# Otherwise print the ballot
else {
    generate_ballot();
}

# end of page
print end_html;
$DB->disconnect;

exit 0;

###############################################################
# get_globals() loads the @CANDIDATES, @CANDIDATE_NAME,
# @PARTIES and @OFFICES globals from information in the
# database.
###############################################################
sub get_globals {

    # fetch the matrix of candidates, parties and offices
    my $query =<<END;
SELECT candidate_id,first_name,last_name,party_name,
       office_name,candidate.party_id,candidate.office_id
  FROM candidate,party,office 
  WHERE candidate.office_id=office.office_id
  AND candidate.party_id=party.party_id
END
;

    my $sth = $DB->prepare($query) or die "Can't prepare: ",$DB->errstr;
    $sth->execute;

    while (my($candidate_id,$first,$last,$party,$office,$party_id,$office_id) 
           = $sth->fetchrow_array) {
        $CANDIDATES[$party_id-1][$office_id-1] = $candidate_id;
        $CANDIDATE_NAME[$candidate_id] = "$first $last";
        $PARTIES[$party_id-1] = $party;
        $OFFICES[$office_id-1] = $office;
    }
    $sth->finish;
}

###############################################################
# generate_ballot(), voting_matrix(), validation_number() and vote()
# create various parts of the page see by the voter
###############################################################
sub generate_ballot {
    print start_multipart_form;
    voting_matrix();
    registration_number();
    vote();
    print end_form;
}

# This generates the table containing the ballot.
sub voting_matrix {
    print img({-src=>'/icons/HandPointing.gif',-align=>'LEFT'}),
      h2('Step 1: Fill in your E-Ballot');

    print
      start_table({-cellspacing => 0,-border=>1}),
      Tr(th(''),th(\@OFFICES));

    for (my $party=0; $party < @PARTIES; $party++) {
        print start_Tr,th($PARTIES[$party]);

        for (my $office=0; $office < @OFFICES; $office++) {

            my $candidate = $CANDIDATES[$party][$office];
            print td({-bgcolor=>$office %2 ? 'white' : '#E0E0E0'},
                      $candidate ? radio_group(-name    => $office,
                                               -value   => $candidate,
                                               -labels  => {$candidate=>$CANDIDATE_NAME[$candidate]},
                                               -default => '-',
                                            )
                                : '&nbsp;'
                    );
        }

        print end_Tr;
    }

    # Handle write-ins.
    print Tr(th('&nbsp;'),
             td([map {radio_group(-name  => $_, -value => 'Write in:').
                       textfield(-name  => "writein $_",
                                -value => '',
                              -override => defined param($_) && param($_)=~/^\d+$/
                   )} (0..$#OFFICES)]
               )
            ),
      end_table;
}



# generate the field for entering voter registration number
sub registration_number {
    print hr,
      img({-src => '/icons/HandPointing.gif', -align => 'LEFT'}),
      h2('Step 2: Enter your Registered Voter Code'),
      blockquote(
                 b('EITHER:'), 'Cut and paste the code here:',
                 textarea(-name => 'registration_id', -rows =>4 , -cols => 70, -wrap => 'physical'), br,
                 b('OR:'), 'Select voter registration file for upload here:', br,
                 filefield(-name => 'registration_file')
               );
}

# generate the VOTE button
sub vote {
    print hr,
      img({-src => '/icons/HandPointing.gif', -align => 'LEFT'}),
    h2('Step 3:','Cast your Ballot'),
    blockquote(b(submit('VOTE')));
}


###############################################################
# validate() validates the ballot to discourage fraud
###############################################################
sub validate {
    # first check that the voter registration field is filled out
    return error('The voter registration ID field is missing.')
      unless param('registration_id') || param('registration_file');

    # check that the voter has voted for at least one office
    return error('The ballot has not been filled out.')
      unless grep {param($_) ne 'Write in:' || param("writein $_")} 0..@OFFICES-1;

    # check that no office has more than one vote
    for (0..@OFFICES-1) {
        my @votes = param($_);
        return error("You have voted for $OFFICES[$_] more than once.") if @votes > 1;
    }

    # recover the registration ID
    my $registration_id;
    if (my $fh = param('registration_file')) {
        while (<$fh>) {
            chomp;
            next unless /--REGISTRATION-START--/../--REGISTRATION-END--/;
            next unless /^\d+$/;
            $registration_id .= $_;
        }
    }
    $registration_id ||= param('registration_id');
    $registration_id =~ s/\D//g;  # get rid of all non-digits
    return error('Your registration code is the incorrect length.')
      unless length $registration_id == REGISTRATION_CODE_LENGTH;

    # check that this is a registered voter
    my $sth = $DB->prepare('SELECT registration_used FROM registration WHERE registration_id=?')
      or die "prepare registration: ",$DB->errstr;
    my $rows = $sth->execute($registration_id);
    return error("The registration code provided is not on the list of eligible voters.")
      unless $rows > 0;

    # check that registration ID has not already been used
    my ($used) = $sth->fetchrow_array;
    return error("That voter registration code has already been used.")
      unless $used == 0;
    $sth->finish;

    return $registration_id;
}

###############################################################
# enter_ballot() updates the database
###############################################################
sub enter_ballot {
    my $registration = shift;

    # lock this registration number so that it can't be used again
    $DB->do("UPDATE registration SET registration_used=1
             WHERE registration_id='$registration'
             AND registration_used=0")>0
      or die "Can't update registration: ",$DB->errstr;

    # generate a ballot ID
    my $id = random_digits(100);

    # prepare the SQL for regular and write-in votes
    my $regular_vote = $DB->prepare("INSERT INTO tally VALUES('$id',?,?,NULL)")
      or die "Can't prepare: ",$DB->errstr;

    my $writein_vote = $DB->prepare("INSERT into writein VALUES('$id',?,?,NULL)")
      or die "Can't prepare: ",$DB->errstr;

    # begin user confirmation
    print h2('Save this Information for your Records');
    print start_dl;
    for my $office (0..$#OFFICES) {
        my $selection      = param($office);
        my $writein        = param("writein $office");
        my $candidate_name = $writein || $CANDIDATE_NAME[$selection] || '-none-';

        # update database with the candidate's vote
        if ($writein) {
            $writein_vote->execute($office+1,$writein)   or die "can't update tally: ", $DB->errstr;
        } elsif ($selection) {
            $regular_vote->execute($office+1,$selection) or die "can't update tally: ", $DB->errstr;
        }

        # update confirmation page
        print dt(b($OFFICES[$office])),dd($candidate_name);
    }
    print end_dl;
    $writein_vote->finish;
    $regular_vote->finish;

    # show user his confirmation number
    $id =~ s/(.{50})/$1\n/;
    print h3('Ballot Confirmation Number'),pre($id);
}

###############################################################
# utilities
###############################################################

# generate some random digits for the ID
sub random_digits {
    my $digits_desired = shift;
    open(RAND, '/dev/urandom') or die "Can't open random number device: $!";
    my $data;
    read(RAND,$data,$digits_desired) or die "Can't read random bytes: $!";
    my @digits = map {$_ % 10} unpack('C*',$data);
    return join '', @digits[0..$digits_desired-1];
}

# all-purpose error message
sub error {
    print p(font({-size=>'+2',-color=>'red'}, @_, br,
                 'Please correct and try again.'));
    return;
}
